import logging
from collections import namedtuple
from typing import List, Optional

from cg.apps.cgstats.db.models import Sample
from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.constants.archiving import ArchiveLocationsInUse
from cg.store import Store

LOG = logging.getLogger(__name__)

PathAndSample = namedtuple("PathAndSample", "path sample_internal_id")


class SpringArchiveAPI:
    """Class handling the archiving of sample SPRING files to an off-premise location for long
    term storage."""

    def __init__(self, housekeeper_api: HousekeeperAPI, status_db: Store):
        self.housekeeper_api: HousekeeperAPI = housekeeper_api
        self.status_db: Store = status_db

    def get_files_by_archive_location(
        self, file_data: List[PathAndSample], archive_location: ArchiveLocationsInUse
    ) -> List[PathAndSample]:
        """
        Returns a list of PathAndSample where the associated sample has a specific archive location.
        """
        selected_files: List[PathAndSample] = []
        for file in file_data:
            sample: Optional[Sample] = self.get_sample(file)
            if sample and sample.archive_location == archive_location:
                selected_files.append(file)
        return selected_files

    def get_sample(self, file: PathAndSample) -> Optional[Sample]:
        sample: Optional[Sample] = self.status_db.get_sample_by_internal_id(file.sample_internal_id)
        if not sample:
            LOG.warning(
                f"No sample found in status_db corresponding to sample_id {file.sample_internal_id}."
                f"Skipping archiving for corresponding file {file.path}."
            )
        return sample



from pathlib import Path
from typing import List

from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.constants.archiving import ArchiveLocationsInUse
from cg.meta.archive.archive import SpringArchiveAPI, PathAndSample
from cg.store.models import Sample


def test_get_files_by_archive_location(
    spring_archive_api: SpringArchiveAPI, populated_housekeeper_api: HousekeeperAPI
):
    """Tests the fetching of sample/customer info from statusdb based on bundle_names
    and returning the samples with the given Archive location."""
    # GIVEN a populated status_db database with two customers, one DDN and one non-DDN,
    # with the DDN customer having two samples, and the non-DDN having one sample.

    # Given non-archived spring files
    non_archived_spring_files: List[PathAndSample] = [
        PathAndSample(path=path, sample_internal_id=sample)
        for sample, path in populated_housekeeper_api.get_non_archived_spring_path_and_bundle_name()
    ]
    # WHEN extracting the files based on data archive
    sorted_spring_files: List[PathAndSample] = spring_archive_api.get_files_by_archive_location(
        non_archived_spring_files, archive_location=ArchiveLocationsInUse.KAROLINSKA_BUCKET
    )

    # THEN there should be spring files
    assert sorted_spring_files
    for file_and_sample in sorted_spring_files:
        sample: Sample = spring_archive_api.status_db.get_sample_by_internal_id(
            file_and_sample.sample_internal_id
        )
        # THEN each file should be correctly sorted on its archive location
        assert sample.customer.data_archive_location == ArchiveLocationsInUse.KAROLINSKA_BUCKET


def test_get_sample_exists(sample_id: str, spring_archive_api: SpringArchiveAPI, spring_file: Path):
    """Tests fetching a sample when the sample exists."""
    # GIVEN a sample that exists in the database
    file_info: PathAndSample = PathAndSample(spring_file, sample_id)

    # WHEN getting the sample
    sample: Sample = spring_archive_api.get_sample(file_info)

    # THEN the correct sample should be returned
    assert sample.internal_id == sample_id


def test_get_sample_not_exists(caplog, spring_archive_api: SpringArchiveAPI, spring_file: Path):
    """Tests fetching a sample when the sample does not exist."""
    # GIVEN a sample that does not exist in the database
    sample_id: str = "non-existent-sample"
    file_info: PathAndSample = PathAndSample(spring_file, sample_id)

    # WHEN getting the sample
    sample: Sample = spring_archive_api.get_sample(file_info)

    # THEN the no sample should be returned
    # THEN both sample_id and file path should be logged
    assert not sample
    assert sample_id in caplog.text
    assert spring_file.as_posix() in caplog.text



from datetime import datetime, timedelta
from pathlib import Path
from typing import List
from unittest import mock

import pytest

from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.constants import SequencingFileTag
from cg.constants.archiving import ArchiveLocationsInUse
from cg.constants.constants import FileFormat
from cg.constants.subject import Gender
from cg.io.controller import WriteStream
from cg.meta.archive.archive import SpringArchiveAPI
from cg.meta.archive.ddn_dataflow import ROOT_TO_TRIM, DDNDataFlowApi, TransferData, TransferPayload
from cg.models.cg_config import DDNDataFlowConfig
from cg.store import Store
from cg.store.models import Customer, Sample
from requests import Response
from tests.store_helpers import StoreHelpers


@pytest.fixture(name="ddn_dataflow_config")
def fixture_ddn_dataflow_config(
    local_storage_repository: str, remote_storage_repository: str
) -> DDNDataFlowConfig:
    """Returns a mock DDN Dataflow config."""
    return DDNDataFlowConfig(
        database_name="test_db",
        user="test_user",
        password="DummyPassword",
        url=Path("some", "api", "url.com").as_posix(),
        local_storage=local_storage_repository,
        archive_repository=remote_storage_repository,
    )


@pytest.fixture(name="ddn_dataflow_api")
def fixture_ddn_dataflow_api(ddn_dataflow_config: DDNDataFlowConfig) -> DDNDataFlowApi:
    """Returns a DDNApi without tokens being set."""
    mock_ddn_auth_success_response = Response()
    mock_ddn_auth_success_response.status_code = 200
    mock_ddn_auth_success_response._content = WriteStream.write_stream_from_content(
        file_format=FileFormat.JSON,
        content={
            "access": "test_auth_token",
            "refresh": "test_refresh_token",
            "expire": (datetime.now() + timedelta(minutes=20)).timestamp(),
        },
    ).encode()
    with mock.patch(
        "cg.meta.archive.ddn_dataflow.APIRequest.api_request_from_content",
        return_value=mock_ddn_auth_success_response,
    ):
        return DDNDataFlowApi(ddn_dataflow_config)


@pytest.fixture(name="transfer_data")
def fixture_transfer_data(local_directory: Path, remote_path: Path) -> TransferData:
    """Return a TransferData object."""
    return TransferData(source=local_directory.as_posix(), destination=remote_path.as_posix())


@pytest.fixture(name="transfer_payload")
def fixture_transfer_payload(transfer_data: TransferData) -> TransferPayload:
    """Return a TransferPayload object containing two identical TransferData object.."""
    return TransferPayload(files_to_transfer=[transfer_data, transfer_data.copy(deep=True)])


@pytest.fixture(name="remote_path")
def fixture_remote_path() -> Path:
    """Returns a mock path."""
    return Path("/some", "place")


@pytest.fixture(name="local_directory")
def fixture_local_directory() -> Path:
    """Returns a mock path with /home as its root."""
    return Path(ROOT_TO_TRIM, "other", "place")


@pytest.fixture(name="trimmed_local_directory")
def fixture_trimmed_local_directory(local_directory: Path) -> Path:
    """Returns the trimmed local directory."""
    return Path(f"/{local_directory.relative_to(ROOT_TO_TRIM)}")


@pytest.fixture(name="local_storage_repository")
def fixture_local_storage_repository() -> str:
    """Returns a local storage repository."""
    return "local@storage:"


@pytest.fixture(name="remote_storage_repository")
def fixture_remote_storage_repository() -> str:
    """Returns a remote storage repository."""
    return "archive@repisitory:"


@pytest.fixture(name="full_remote_path")
def fixture_full_remote_path(remote_storage_repository: str, remote_path: Path) -> str:
    """Returns the merged remote repository and path."""
    return remote_storage_repository + remote_path.as_posix()


@pytest.fixture(name="full_local_path")
def fixture_full_local_path(local_storage_repository: str, trimmed_local_directory: Path) -> str:
    """Returns the merged local repository and trimmed path."""
    return local_storage_repository + trimmed_local_directory.as_posix()


@pytest.fixture(name="archive_store")
def fixture_archive_store(
    base_store: Store,
    helpers: StoreHelpers,
    sample_id,
    father_sample_id,
    mother_sample_id,
    sample_name,
) -> Store:
    """Returns a store with samples for both a DDN customer as well as a non-DDN customer."""
    customer_ddn: Customer = base_store.add_customer(
        internal_id="CustWithDDN",
        invoice_address="Baker Street 221B",
        invoice_reference="Sherlock Holmes",
        name="Sherlock Holmes",
        is_clinical=True,
        data_archive_location=ArchiveLocationsInUse.KAROLINSKA_BUCKET,
    )
    customer_without_ddn: Customer = base_store.add_customer(
        internal_id="CustWithoutDDN",
        invoice_address="Wallaby Way 42",
        invoice_reference="P Sherman",
        name="P Sherman",
        is_clinical=False,
        data_archive_location="PDC",
    )

    new_samples: List[Sample] = [
        base_store.add_sample(
            name=sample_name,
            sex=Gender.MALE,
            internal_id=sample_id,
        ),
        base_store.add_sample(
            name="sample_2_with_ddn_customer",
            sex=Gender.MALE,
            internal_id=mother_sample_id,
        ),
        base_store.add_sample(
            name="sample_without_ddn_customer",
            sex=Gender.MALE,
            internal_id=father_sample_id,
        ),
    ]
    new_samples[0].customer = customer_ddn
    new_samples[1].customer = customer_ddn
    new_samples[2].customer = customer_without_ddn
    external_app = base_store.get_application_by_tag("WGXCUSC000").versions[0]
    wgs_app = base_store.get_application_by_tag("WGSPCFC030").versions[0]
    for sample in new_samples:
        sample.application_version = external_app if "external" in sample.name else wgs_app
    base_store.session.add(customer_ddn)
    base_store.session.add(customer_without_ddn)
    base_store.session.add_all(new_samples)
    base_store.session.commit()
    return base_store


@pytest.fixture(name="spring_archive_api")
def fixture_spring_archive_api(
    populated_housekeeper_api: HousekeeperAPI,
    archive_store: Store,
    father_sample_id: str,
    helpers,
) -> SpringArchiveAPI:
    """Returns an ArchiveAPI with a populated housekeeper store and a DDNDataFlowApi.
    Also adds /home/ as a prefix for each SPRING file for the DDNDataFlowApi to accept them."""
    for spring_file in populated_housekeeper_api.files(tags=[SequencingFileTag.SPRING]):
        spring_file.path = f"/home/{spring_file.path}"
    populated_housekeeper_api.commit()
    return SpringArchiveAPI(
        housekeeper_api=populated_housekeeper_api,
        status_db=archive_store,
    )
